"""TTY-rendering enabled multi progress bar."""
from __future__ import annotations

from dataclasses import InitVar, dataclass, field
from typing import List, Optional, Tuple

from ..progress import BarInfo, LogicalProgressBar
from .console import at_position
from .drawing_mixin import DrawingWrapperMixin
from .position import Position
from .tty_drawable import TTYDrawableMixin


@dataclass
class _TTYMultiProgressBarData(TTYDrawableMixin['TTYMultiProgressBar']):
    """Mypy crashes if this is part of TTYMultiProgressBar yet."""

    bar_infos: InitVar[Optional[List[BarInfo]]] = None

    _bars: List[Tuple[Position, LogicalProgressBar]] = field(
        default_factory=lambda: [], init=False
    )
    _started: bool = field(default=False, init=False)

    def __post_init__(self, bar_infos) -> None:
        bar_infos = bar_infos if bar_infos is not None else []

        for delta, bar in enumerate(bar_infos):
            self._bars.append((Position(delta, 0), LogicalProgressBar(bar)))


@dataclass
class _MultiProgressBarProxyData:
    bar: LogicalProgressBar
    parent: TTYMultiProgressBar


@dataclass
class MultiProgressBarProxy(DrawingWrapperMixin, _MultiProgressBarProxyData):
    """Proxy for a multi-progress single bar capable of drawing its parent."""

    def draw(self):
        """Draw the parent TTYMultiProgressBar."""
        self.parent.draw()


class TTYMultiProgressBar(_TTYMultiProgressBarData):
    """Progress bar capable of "drawing" itself to an ANSI-escape enabled TTY.

    Args:
        bar_infos: Parameters for initializing the inner progress bars.
        position: The position where the progress bar should be rendered.
                  If ommitted, uses the current cursor position.
    """

    @property
    def _height(self):
        return len(self._bars)

    def _draw(self):
        """Draw the progress bar to the tty."""
        for delta, bar in self._bars:
            with at_position(self.position + delta):
                print(bar.render(), end='', flush=True)

    def bar_at(self, index: int) -> MultiProgressBarProxy:
        """Get a proxy to the bar at a given index."""
        _, bar = self._bars[index]

        return MultiProgressBarProxy(bar=bar, parent=self)


__all__ = [
    'MultiProgressBarProxy',
    'TTYMultiProgressBar',
]



"""Main logic behind progress bar rendering."""
from dataclasses import dataclass, field

from .error import ProgressValueError
from .util import checked_property, wrapper


def _positive(name, instance, value: int):
    if value <= 0:
        raise ProgressValueError(
            f'Value for {name} must be positive. Attempted value: {value}.'
        )


def _non_negative(name, instance, value: int):
    if value < 0:
        raise ProgressValueError(
            f'Value for {name} must be non-negative. Attempted value: {value}.'
        )


def _lower_than_total(_, instance, value: int):
    if not 0 <= value <= instance.total:
        raise ProgressValueError(
            f'Value for progress must be between 0 and {instance.total}. '
            f'Attempted value was {value}.'
        )


def _bigger_than_progress(_, instance, value: int):
    try:
        if value < instance.progress:
            raise ProgressValueError(
                f'Value for total cannot lower than current '
                f'progress: {instance.progress}. '
                f'Attempted value was {value}.'
            )
    except KeyError:
        pass


def _more_than_two_characters(_, instance, value: str):
    if len(value) < 2:
        raise ProgressValueError(
            '"characters" must have at least two characters. '
            'At least an empty and a full character must be provided. '
            f'Attempted value: {repr(value)}.'
        )


@dataclass
class BarInfo:
    """Internal progress bar data."""

    size: int = checked_property(validators=(_non_negative,))
    total: int = checked_property(validators=(_non_negative, _bigger_than_progress),)
    progress: int = checked_property(default=0, validators=(_lower_than_total,))
    characters: str = checked_property(
        default=' ⠁⠃⠇⡇⣇⣧⣷⣿', validators=(_more_than_two_characters,)
    )


@dataclass
class LogicalProgressBar:
    r"""A text-based progress bar.

    The progress bar is composed by three parts, as seen below in between '|':

    |⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣧                      |
      \    solid    / | \        empty     /
                    leading

    The solid part represents everything in the progress bar that won't change
    anymore.

    The leading character can progress through a sequence of symbols,
    artificially raising the progress bar resolution.

    The empty part is composed of characters that represent progress that
    hasn't yet been reached.

    Args:
        size: The progress bar size in characters.
        total: The total progress capacity of the bar in arbitrary units.
        progress: The starting progress of the progress bar.
        characters: The characters to use for the bar. The first one is the
                    empty one, the last one is the solid one. Everything in
                    between represents the progression of the leading
                    character, from left to right.
    """
    bar_info: BarInfo
    _steps: int = field(init=False, repr=False)
    _resolution: int = field(init=False, repr=False)

    size: int = wrapper(lambda self: self.bar_info)
    total: int = wrapper(lambda self: self.bar_info)
    progress: int = wrapper(lambda self: self.bar_info)
    characters: str = wrapper(lambda self: self.bar_info)

    def _steps_and_resolution(self):
        steps = len(self.characters) - 1
        resolution = self.size * (steps)

        return steps, resolution

    def __post_init__(self):
        self._steps, self._resolution = self._steps_and_resolution()

    def _filled_amount(self):
        full_ratio = self.progress / self.total

        return int(self._resolution * full_ratio)

    def _solid_size(self, filled):
        return max(0, filled - 1) // self._steps

    def _leading_char_index(self, filled):
        if filled == 0:
            return 0

        return (filled - 1) % self._steps + 1

    def _solid_and_leading(self):
        filled = self._filled_amount()

        solid = self._solid_size(filled) * self.characters[-1]
        leading = self.characters[self._leading_char_index(filled)]

        return solid, leading

    def render(self) -> str:
        """Render the progress bar to a string.

        Returns:
            The rendered progress bar.
        """
        solid, leading = self._solid_and_leading()
        empty = self.characters[0]
        full = f'{solid}{leading}'
        bar = f'{full:{empty}<{self.size}}'

        return bar


__all__ = [
    'BarInfo',
    'LogicalProgressBar',
]



"""TTY-rendering enabled progress bar."""
from dataclasses import InitVar, dataclass, field

from ..progress import BarInfo, LogicalProgressBar
from .console import at_position
from .drawing_mixin import DrawingWrapperMixin
from .tty_drawable import TTYDrawableMixin


@dataclass
class _TTYProgressBarData:
    bar_info: InitVar[BarInfo]


@dataclass
class TTYProgressBar(
    TTYDrawableMixin['TTYProgressBar'], DrawingWrapperMixin, _TTYProgressBarData,
):
    """Progress bar capable of "drawing" itself to an ANSI-escape enabled TTY.

    Args:
        position: The position where the progress bar should be rendered.
                  If ommitted, uses the current cursor position.
    """

    bar: LogicalProgressBar = field(init=False)
    _height = 1

    def __post_init__(self, bar_info):
        self.bar = LogicalProgressBar(bar_info)

    def _draw(self):
        """Draw the progress bar to the tty."""
        with at_position(self.position):
            print(self.bar.render(), end='', flush=True)


__all__ = ['TTYProgressBar']



"""Mixin for objects that wrap a progress bar and draw a drawable."""
from dataclasses import dataclass
from typing_extensions import Protocol

from ..types import ProgressBar
from ..util import wrapper


class DrawableWithBar(Protocol):
    """A class that has a ProgressBar and can draw itself."""

    bar: ProgressBar

    def draw(self):
        """Draw self somewhere."""
        ...


@dataclass
class DrawingWrapperMixin:
    """Proxy for a multi-progress single bar capable of drawing its parent."""

    size: int = wrapper(lambda self: self.bar)
    total: int = wrapper(lambda self: self.bar)
    characters: str = wrapper(lambda self: self.bar)

    @property
    def progress(self: DrawableWithBar) -> int:
        """Wrap inner bar progress by drawing when it is set."""
        return self.bar.progress

    @progress.setter
    def progress(self: DrawableWithBar, value: int):
        self.bar.progress = value
        self.draw()  # type: ignore



"""Utility types and functions."""
from typing import Any, Callable, Optional, Sequence, TypeVar

T = TypeVar('T')
Validator = Callable[[Optional[str], Any, T], bool]


class CheckedProperty:
    """Helper to have a validated member in a dataclass.

    Dataclasses don't really play very well with properties, so this works
    around that issue.

    Args:
        validators: Predicates to run agains a value when set.
        default: Default value for the property.
    """

    def __init__(
        self, validators: Sequence[Validator[T]] = (), default: Optional[T] = None
    ):
        self.validators = validators
        self.default = default
        self.name = None

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if not instance:
            return self
        return instance.__dict__[self.name]

    def __delete__(self, instance):
        del instance.__dict__[self.name]

    def __set__(self, instance, value: T):
        if value is self:
            if self.default is not None:
                value = self.default
            else:
                type_name = type(instance).__name__
                raise ValueError(
                    f'Attribute "{self.name}" on type {type_name} has '
                    'no default value and none was provided.'
                )

        for validator in self.validators:
            validator(self.name, instance, value)
        instance.__dict__[self.name] = value


def checked_property(*args, **kwargs) -> Any:
    """Type-erased constructor for CheckedProperty."""
    return CheckedProperty(*args, **kwargs)


class Wrapper:
    """Helper to wrap a member's attributes.

    Args:
        wrapped_getter:
    """

    def __init__(self, wrapped_getter):
        self.wrapped_getter = wrapped_getter
        self.name = None

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if not instance:
            return self
        return getattr(self.wrapped_getter(instance), self.name)

    def __delete__(self, instance):
        delattr(self.wrapped_getter(instance), self.name)

    def __set__(self, instance, value):
        if value is self:
            return

        setattr(self.wrapped_getter(instance), self.name, value)


def wrapper(*args, **kwargs) -> Any:
    """Type-erased constructor for CheckedProperty."""
    return Wrapper(*args, **kwargs)



"""Simple vocabulary types and protocols."""
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class ProgressBar(Protocol):
    """Protocol of every ProgressBar implementation."""

    size: int
    progress: int
    total: int
    characters: str



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['proggy', 'proggy.tty']

package_data = \
{'': ['*']}

install_requires = \
['typing_extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'proggy',
    'version': '0.4.3',
    'description': 'A progress bar generator.',
    'long_description': "Proggy\n======\n\nProgressively progressing through progress bar generation.\n\nProggy generates text-based progress bars. Mildly inspired by Rust's\n[indicatif](https://github.com/mitsuhiko/indicatif).\n\nProggy only renders progress bars to a string. Displaying them is, as of now,\nnot handled and left to the user.\n\nExamples\n--------\n\n### API usage\n\n```\n>>> from proggy import ProgressBar\n>>> pb = ProgressBar(30, 100, progress=75)\n>>> pb.render()\n'⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⣿⡇       '\n```\n\n### CLI output\n\n#### Single progress bar (`test.py`):\n```python\nimport time\n\nfrom proggy import BarInfo\nfrom proggy.tty import TTYProgressBar\n\n\nwith TTYProgressBar(BarInfo(size=30, total=100)) as p:\n    for i in range(100):\n        time.sleep(0.1)\n        p.progress += 1\n```\n\nOutput:\n\n![test.py output](gif/test.gif)\n\n#### Multi-progress bar with `asyncio` (`async_test.py`):\n```\nfrom asyncio import ensure_future, run, sleep\n\nfrom proggy import BarInfo, LogicalProgressBar\nfrom proggy.types import ProgressBar\nfrom proggy.tty import TTYMultiProgressBar\n\n\nasync def task1(p: ProgressBar):\n    for _ in range(100):\n        await sleep(0.1)\n        p.progress += 1\n\n\nasync def task2(p: ProgressBar):\n    for _ in range(50):\n        await sleep(0.1)\n        p.progress += 1\n\n\nasync def main():\n    with TTYMultiProgressBar(bar_infos=[\n        BarInfo(size=30, total=100),\n        BarInfo(size=30, total=50),\n    ]) as mp:\n        x = ensure_future(task1(mp.bar_at(0)))\n        await task2(mp.bar_at(1))\n        await x\n\nrun(main())\n```\n\nOutput:\n\n![async_test.py output](gif/async_test.gif)\n",
    'author': 'Tarcísio Eduardo Moreira Crocomo',
    'author_email': 'tarcisioe@pm.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Console and cursor control."""
from contextlib import contextmanager
from sys import stdin
from termios import TCSANOW, tcgetattr, tcsetattr
from tty import setcbreak
from typing import Optional

from .ansi import (
    device_status_report_escape,
    read_ansi_response,
    set_cursor_position_escape,
)
from .position import Position


@contextmanager
def cbreak():
    """Set cbreak on TTY and reset after."""
    attr = tcgetattr(stdin)
    setcbreak(stdin, TCSANOW)
    yield
    tcsetattr(stdin, TCSANOW, attr)


def try_get_cursor_position() -> Optional[Position]:
    """Try to get the cursor position.

    Can fail if there was user input during execution.

    Returns:
        On success, the current cursor position, else None.
    """
    with cbreak():
        print(device_status_report_escape, end='', flush=True)
        y, x = read_ansi_response(stdin.buffer)

    try:
        return Position(int(y), int(x))
    except ValueError:
        return None


def get_cursor_position() -> Position:
    """Get the current cursor position in the TTY.

    Returns:
        The current cursor position.
    """
    position = None

    while position is None:
        position = try_get_cursor_position()

    return position


def set_cursor_position(position: Position):
    """Set the cursor to a given position in the TTY.

    Args:
        position: Where to set the cursor.
    """
    print(set_cursor_position_escape(position), end='', flush=True)


@contextmanager
def at_position(position: Position):
    """Set the cursor position to a given position and move it back after.

    Args:
        position: Where to set the cursor.
    """
    old = get_cursor_position()
    set_cursor_position(position)
    yield
    set_cursor_position(old)


def reserve_space(
    n_lines: int, starting_position: Optional[Position] = None,
) -> Position:
    """Reserve n lines in the console and return the position of the first one.

    Leaves the cursor in the last reserve line.

    Args:
        n_lines: How many lines to reserve.
        starting_position: How many lines to reserve.

    Returns:
        The new starting position (which may have changed if scrolling
        occurred).
    """
    start = (
        starting_position if starting_position is not None else get_cursor_position()
    )

    set_cursor_position(start)
    print('\n' * (n_lines - 1), end='', flush=True)
    ending_line_position = get_cursor_position() - Position(n_lines - 1, 0)
    return Position(ending_line_position.y, start.x)


__all__ = [
    'at_position',
    'get_cursor_position',
    'set_cursor_position',
    'reserve_space',
]



"""TTY-drawable base class."""
from dataclasses import dataclass, field
from typing import Generic, Optional, TypeVar, cast

from .console import get_cursor_position, reserve_space
from .error import TTYDrawableNotStartedError
from .position import Position

T = TypeVar('T')


@dataclass
class TTYDrawableMixin(Generic[T]):
    """Abstract TTY-drawable context manager.

    Child classes should provide their own height and method of drawing.
    """

    position: Optional[Position] = None
    _started: bool = field(default=False, init=False)
    _height: int = field(init=False)

    def draw(self) -> None:
        """Draw to the TTY."""
        if not self._started:
            raise TTYDrawableNotStartedError(
                f'{type(self).__name__} objects should only be used as '
                'context managers.'
            )

        self._draw()  # type: ignore

    def __enter__(self) -> T:
        self._started = True

        if self.position is None:
            self.position = get_cursor_position()

        self.position = reserve_space(self._height, starting_position=self.position)

        self.draw()
        return cast(T, self)

    def __exit__(self, exc_type, exc, exc_tb):
        print()
        self._started = False


__all__ = ['TTYDrawableMixin']



"""TTY-related functionality package."""
# pylint: skip-file
# flake8: noqa

from .console import *
from .position import *
from .tty_multi_progress_bar import *
from .tty_progress_bar import *

__all__ = [
    *console.__all__,  # type: ignore
    *position.__all__,  # type: ignore
    *tty_multi_progress_bar.__all__,  # type: ignore
    *tty_progress_bar.__all__,  # type: ignore
]
